"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
exports.axios = exports.devMode = exports.solutionFinder = exports.AuthConfig = exports.Config = exports.version = exports.SessionSecret = exports.repoLogger = exports.mainLogger = void 0;
const solutionfinder_1 = require("./solutionfinder");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const winston_1 = __importDefault(require("winston"));
const config_reader_1 = require("./utils/config-reader");
const axios_1 = __importDefault(require("axios"));
const https_1 = __importDefault(require("https"));
const crypto_1 = __importDefault(require("crypto"));
///
/// Multiple imports will still cause this code to only be executed once.
/// The exports here will not be duplicated. They will be created only once and 
/// have the same handles even when imported in different files. 
/// This is due to the Typescript ES2020 import decoration
/// 
/// 
// Create all of the required directories.
const requiredDirectories = ["log", "tmp", "data"];
requiredDirectories.forEach(dir => {
    if (!fs_1.default.existsSync(__dirname + "/../" + dir))
        fs_1.default.mkdirSync(__dirname + '/../' + dir);
});
exports.mainLogger = winston_1.default.createLogger({
    level: 'info',
    format: winston_1.default.format.combine(winston_1.default.format.label({ label: 'Main' }), winston_1.default.format.timestamp({ format: 'YYYY-MM-DD HH:mm:ss.SSS' }), winston_1.default.format.printf(({ level, message, label, timestamp }) => {
        return `${timestamp} [${label}] ${level}: ${message}`;
    })),
    transports: [
        new winston_1.default.transports.File({ filename: __dirname + '/../log/error.log', level: 'error', maxsize: 10 * 1024 * 1024, maxFiles: 1 }),
        new winston_1.default.transports.File({ filename: __dirname + '/../log/verbose.log', maxsize: 10 * 1024 * 1024, maxFiles: 1 }),
    ],
});
// Store only 1MB of entries.
exports.repoLogger = winston_1.default.createLogger({
    level: 'info',
    format: winston_1.default.format.combine(winston_1.default.format.label({ label: 'Repo' }), winston_1.default.format.timestamp({ format: 'YYYY-MM-DD HH:mm:ss.SSS' }), winston_1.default.format.printf(({ level, message, label, timestamp }) => {
        return `${timestamp} [${label}] ${level}: ${message}`;
    })),
    transports: [
        new winston_1.default.transports.File({ filename: __dirname + '/../log/repo.log', maxsize: 1 * 1024 * 1024, maxFiles: 1 })
    ],
});
// Create a configuration reader object that will restart the process on
// any config file changing.
const configReader = new config_reader_1.ConfigReader();
configReader.on('error', (msg, err) => {
    exports.mainLogger.error(msg + ": " + err);
});
configReader.on('change', (file, data) => {
    console.error("Configuration file '" + file + "' changed. Killing process...");
    process.exitCode = 0;
    // Give it up to 15 seconds to close.
    setTimeout(process.exit, 15000);
});
let sessionSecret;
const sessionFile = __dirname + `/../data/session.secret`;
if (fs_1.default.existsSync(sessionFile))
    sessionSecret = fs_1.default.readFileSync(sessionFile, { encoding: "utf-8" });
else {
    sessionSecret = crypto_1.default.randomBytes(32).toString();
    fs_1.default.writeFileSync(sessionFile, sessionSecret);
}
exports.SessionSecret = sessionSecret;
// Starting from the ./esa-solution-server/server directory: 
const agPluginDeployDir = path_1.default.normalize(__dirname + "/../../");
const agRemotePluginDir = path_1.default.normalize(agPluginDeployDir + "../");
const agAgentConfDir = path_1.default.normalize(agRemotePluginDir + "agent/conf");
// Read ActiveGate config.
const ruxit = configReader.loadFile(agAgentConfDir + '/ruxitagent.conf', false) || {};
// When running, if we're not in an ActiveGate or we can't find this data,
// we will default to use a configuration specified in the main Config.
if (!ruxit.Tenant)
    console.warn("No tenant ID detected.");
// Read configuration files
const npmConfig = configReader.loadFile(__dirname + '/package.json');
exports.version = npmConfig.version;
exports.Config = configReader.loadFile(__dirname + '/../dss.config.json') || npmConfig['esa-solution-server'];
exports.AuthConfig = configReader.loadFile(__dirname + '/../config/authorizations.json') ||
    {
        uiLoginMode: 'dynatrace',
        dynatrace: { endpoint: `https://127.0.0.1:9999/e/${ruxit.Tenant}/` }
    };
if (exports.AuthConfig.openid) // Set secret for OpenID.
    exports.AuthConfig.openid.secret = exports.AuthConfig.openid.secret || sessionSecret;
const certs = configReader.loadFile(__dirname + '/../config/certificates.json') || {};
// Explicitly set properties in main configuration object
exports.Config.name = npmConfig.name;
exports.Config.version = npmConfig.version;
exports.Config.tenantId = ruxit === null || ruxit === void 0 ? void 0 : ruxit.Tenant;
// Create the Dynatrace endpoint connection string based on a supplied configuration
// templating the tenant ID or try connecting locally to the AG on 9999.
exports.Config.dtUrl = exports.Config.tenantId
    ? (((_a = exports.Config.dtUrl) === null || _a === void 0 ? void 0 : _a.replace("{{tenant}}", exports.Config.tenantId))
        || `https://127.0.0.1:9999/e/${exports.Config.tenantId}/`)
    : exports.Config.dtUrl;
exports.Config.token = exports.AuthConfig.altToken;
exports.Config.certs = certs;
exports.Config.dssRoot = __dirname + '/../';
// Create an object containing user permissions for the 
// Authorization configuration
let authorizationMap = {};
(_b = exports.AuthConfig.authorizationSetup) === null || _b === void 0 ? void 0 : _b.forEach(scope => {
    authorizationMap[scope.userId] = scope.permissions;
});
exports.AuthConfig.authorizations = authorizationMap;
// This acts as our repo of solutions and extensions.
// It does a periodic sweep to discover/update/remove them.
exports.solutionFinder = globalThis.solutionFinder = new solutionfinder_1.SolutionFinder(1 * 60 * 1000); // TODO: Take from config.
exports.solutionFinder.start();
// Check if the app is running in "prod" mode.
// AKA running as a plugin with fully compiled JS.
exports.devMode = /\.ts$/.test(process.argv[1]) || process.argv[2] == '--watch';
exports.axios = axios_1.default.create({
    httpsAgent: new https_1.default.Agent({
        rejectUnauthorized: false
    }),
    proxy: exports.AuthConfig.proxy
});
