"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SessionStoreClient = exports.SharedSessionStore = void 0;
const axios_1 = __importDefault(require("axios"));
const express_1 = __importDefault(require("express"));
const level_1 = __importDefault(require("level"));
const memory_level_1 = require("memory-level");
const http_1 = __importDefault(require("http"));
/**
 * Run on "primary"
 */
class SharedSessionStore {
    /**
     *
     * useful if you intend to use multiple session stores.
     * @param port port that the store will listen on.
     */
    constructor(port = 6801, options) {
        // If we are storing the sessions.
        this.db = {};
        if (options) {
            this.createInstance(options);
        }
        else
            this.createInstance({ instanceId: "default" });
        this.launchServer(port);
    }
    createInstance(options) {
        const { cacheOnDisk, cachePath, maxStoreTime, instanceId } = options;
        const levelOpts = { valueEncoding: 'json' };
        // Create a stored level instance if a cache path is defined.
        // Otherwise we just store all session in memory.
        let db = cacheOnDisk
            ? (0, level_1.default)(cachePath, levelOpts)
            : new memory_level_1.MemoryLevel(levelOpts);
        if (this.db[instanceId])
            throw Error("Database with id " + instanceId + " already exists!");
        this.db[instanceId] = db;
    }
    /**
     * Method to release all handles and flush all writes to disk.
     * If the program terminates without this called, in rare cases data could be lost.
     */
    destroy() {
        // Close all open databases. This flushes their contents to disk.
        // Of course, this doesn't do anything for memory db mode.
        Object.keys(this.db).forEach(store => this.db[store].close());
    }
    launchServer(port) {
        const app = this.app = (0, express_1.default)();
        app.use(express_1.default.json());
        // Handle incoming connections.
        app.use((req, res, next) => {
            const { storeId, action, key, value } = req.body;
            if (!action || !key || (action == "put" && !value) || (action == "batch" && !value))
                return next({ status: 400, message: "Bad request" });
            const db = this.db[storeId || "default"];
            const storeVal = {
                storeTime: new Date().getTime(),
                value: value
            };
            // console.log(storeVal);
            // TBD batch
            console.log(`db[${action}](${key || value}, ${key ? storeVal : null})`);
            const p = action == "get" ? db.get(key) :
                action == "put" ? db.put(key, storeVal) :
                    action == "del" ? db.del(key) :
                        db.batch(value);
            // If we don't have a key, directly pass in data
            p.then(data => {
                // Found data in store. we return subitem `value` from stored value.
                // This allows us to add extra attributes to the data without polluting the 
                // response.
                // Return data or 'true' if the action was successful.
                // console.log("Sending response", data);
                res.send({
                    value: data ? data.value : true
                });
            }).catch(err => {
                // Missing data in store. We return `undefined`.
                if (err.code == "LEVEL_NOT_FOUND" || err.type == "NotFoundError")
                    return res.send(undefined);
                // If there is some other error -- pass it back to the client.
                next(err);
            });
        });
        // Unmatched routes are a 404.
        app.use((req, res, next) => next({ status: 404, message: `Not Found: ${req.url}` }));
        // Handle any exceptions.
        app.use((err, req, res, next) => {
            const error = err.stack || err.message || err.toString();
            // Don't bother logging 404 errors in the general log.
            if (err.status != 404)
                console.error(error);
            // Render the error page.
            res.status(err.status || 500);
            // If the browser is accepting HTML in the request, we'll render a nice error page.
            // Otherwise: JSON.
            res.send(error);
        });
        const server = this.server = http_1.default.createServer(app);
        server.listen(port);
        server.on('error', (error) => {
            if (error.syscall !== 'listen')
                throw error;
            // Report friendly errors where we can.
            switch (error.code) {
                case 'EACCES':
                    console.error('Port ' + port + ' requires elevated privileges');
                    process.exit(1);
                case 'EADDRINUSE':
                    console.error('Port ' + port + ' is already in use');
                    process.exit(1);
                default:
                    throw error;
            }
        });
        server.on('listening', () => {
            const addr = server.address();
            console.log("Server %s listening on %s.", process.pid, addr.port || addr);
        });
    }
}
exports.SharedSessionStore = SharedSessionStore;
/**
 * Run in worker instances
 */
class SessionStoreClient {
    /**
     * Create a new Session Store Client instance.
     * @param port Port of Level cluster to listen on
     * @param instanceId Instance ID for this client. If you have multiple instances that you need to interface, create multiple client objects.
     */
    constructor(port = 6801, instanceId = "default") {
        this.port = port;
        this.instanceId = instanceId;
    }
    /**
     * Get an item from the store
     * @param key Key to look for item at
     * @returns stored item OR `undefined`. Getting a response of `undefined` means that there was no value for the given key.
     *
     */
    get(key) {
        return this.request(this.instanceId, key, "get");
    }
    /**
     * Put an item into the store.
     * @param key   Key to store value at.
     * @param value Value to store at key.
     * @returns true/false. Returns true if operation was successful.
     */
    put(key, value) {
        return this.request(this.instanceId, key, "put", value);
    }
    /**
     * Delete an item from the store.
     * @param key Key of item to delete
     * @returns true/false. Returns true if operation was successful.
     */
    delete(key) {
        return this.request(this.instanceId, key, "del");
    }
    /**
     * @experimental
     * @param entries
     * @returns
     */
    batch(entries) {
        return this.request(this.instanceId, null, "batch", entries);
    }
    /**
     * Make the request to the target store.
     *
     * @param storeId string ID of cache within store.
     * @param key     which element in the store to work on
     * @param action  "get" "put" "del" -- do we delete the key, set the value for the key, or get the value?
     * @param value   when action is "put", what to store for the specified key.
     * @returns
     */
    request(storeId, key, action, value) {
        var _a, _b, _c, _d;
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let { data } = yield axios_1.default.put(`http://127.0.0.1:${this.port}/`, {
                    storeId,
                    key,
                    action,
                    value
                });
                return data.value;
            }
            catch (ex) {
                // Something horrible happened.
                // TODO: throw different error if target server is not listening
                return {
                    status: 500,
                    message: "Internal Failure.",
                    ex: {
                        target: (_a = ex.config) === null || _a === void 0 ? void 0 : _a.url,
                        payload: (_b = ex.config) === null || _b === void 0 ? void 0 : _b.data,
                        data: (_c = ex.response) === null || _c === void 0 ? void 0 : _c.data,
                        status: (_d = ex.response) === null || _d === void 0 ? void 0 : _d.status
                    }
                };
            }
        });
    }
}
exports.SessionStoreClient = SessionStoreClient;
