"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolutionFinder = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const shelljs_1 = __importDefault(require("shelljs"));
const express_1 = __importDefault(require("express"));
const express_http_proxy_1 = __importDefault(require("express-http-proxy"));
const serve_index_1 = __importDefault(require("serve-index"));
const errorhandler_1 = require("./utils/errorhandler");
const utils_1 = require("./utils/utils");
const globals_1 = require("./globals");
const serveWithStylesheet = "./static/serveindex.css";
class SolutionFinder {
    constructor(discoveryInterval = 60 * 1000) {
        this.discoveryInterval = discoveryInterval;
        this.dssRootDir = globals_1.Config.dssRoot;
        this.parentDir = globals_1.Config.dssRoot + '../';
        this.solutions = [];
        this.extensions = [];
        this.blink = 0;
        this.timer = null;
        // Project directory root. Detected as parent of current directory (server).
        this.runningDirectory = path_1.default.parse(path_1.default.parse(__dirname).dir).base;
        this.discoveryInterval = discoveryInterval;
        this.logger = globals_1.repoLogger;
    }
    isConfigValid(config, solutionDir) {
        // Create a comma delimited list with missing properties.
        const missingProperties = [
            (config.name == undefined) && "name",
            (config.path == undefined) && "path",
            (config.title == undefined) && "title",
            // Port mode requires some properties.
            (!["proxy", "static", "site"].includes(config.mode)) && "mode",
            // Port mode requires some properties.
            (config.mode == 'proxy' && (config.protocol == undefined)) && "protocol",
            (config.mode == 'proxy' && (config.port == undefined)) && "port",
        ]
            .filter(f => typeof f === 'string');
        if (missingProperties.length > 0) {
            config.status = "configRequired";
            config.missingProps = missingProperties.join(', ');
            this.logger.warn(`    Solution configuration is missing ` + missingProperties.join(', '));
        }
        else {
            if (!config.image) {
                config.image = '/images/plugin.svg';
            }
            else if (!config.image.startsWith('http')) {
                config.image = path_1.default.normalize(config.path + '/' + config.image);
            }
            // Does the solution have a doc folder?
            config.hasDocs = false;
            if (fs_1.default.existsSync(solutionDir + '/doc')) {
                const docs = shelljs_1.default.ls(solutionDir + '/doc');
                config.hasDocs = docs.length > 0;
            }
            // Does the solution have any log files?
            config.hasLogs = false;
            if (fs_1.default.existsSync(solutionDir + '/log')) {
                const logs = shelljs_1.default.ls(solutionDir + '/log')
                    .filter(f => /\.(log|txt)$/.test(f));
                config.hasLogs = logs.length > 0;
            }
            // Does the solution have any config files in /config?
            config.hasConfigs = false;
            if (fs_1.default.existsSync(solutionDir + '/config')) {
                const configs = shelljs_1.default.ls(solutionDir + '/config')
                    .filter(f => /\.(json|ini|yml|xml|properties)$/.test(f));
                config.hasConfigs = configs.length > 0;
            }
            // Does solution have an ecosystem config
            config.hasEcosystem = false;
            if (fs_1.default.existsSync(solutionDir + '/ecosystem.config.js')) {
                config.hasEcosystem = true;
            }
        }
        return missingProperties.length === 0;
    }
    hasConfigChanged(solution, configJSON) {
        const changedProperties = [
            (configJSON.title !== solution.config.title) && "title",
            (configJSON.path !== solution.config.path) && "path",
            (configJSON.port !== solution.config.port) && "port",
            (configJSON.image !== solution.config.image) && "image",
            (configJSON.protocol !== solution.config.protocol) && "protocol",
            (configJSON.description !== solution.config.description) && "description"
        ]
            .filter(f => typeof f === 'string');
        if (changedProperties.length > 0) {
            this.logger.warn(`    Solution configuration changed ` + changedProperties.join(', '));
        }
        return changedProperties.length > 0;
    }
    createRouter(config, solutionDir) {
        let router = null;
        let handler = null;
        try {
            switch (config.mode) {
                case "proxy":
                    this.logger.info(`    Loading solution in 'proxy' mode at port ${config.port}`);
                    handler = (0, express_http_proxy_1.default)((config.protocol || 'https') + '://127.0.0.1:' + config.port, {
                        preserveHostHdr: true,
                        limit: '50mb',
                        proxyReqOptDecorator: (opts, req) => {
                            opts.rejectUnauthorized = false; // Allow self-signed certificates. 
                            opts.rewritePath = true;
                            opts.headers = req.headers;
                            opts.headers['X-Forwarded-For'] = (req.headers['X-Forwarded-For']
                                ? req.headers['X-Forwarded-For']
                                : req.ip) + ", 127.0.0.1";
                            return opts;
                        },
                        proxyErrorHandler: (err, res, next) => {
                            console.info("Proxy error", (config.protocol || 'https') + '://127.0.0.1:' + config.port);
                            console.error(err);
                            // TODO: Provide clean UI for services that aren't responding.
                            switch (err === null || err === void 0 ? void 0 : err.code) {
                                case 'ECONNRESET':
                                    // The other side of the TCP conversation abruptly closed its end 
                                    // of the connection. This is most probably due to a protocol error. 
                                    return res.status(405).send("Connection Reset");
                                case 'ECONNREFUSED': {
                                    return res.status(404).send((0, errorhandler_1.render404Page)(res.req, { title: "Solution Refused request.", message: `Unable to connect to solution at ${res.req.method}::${res.req.originalUrl}<br/>Check if your solution is running and healthy.` }));
                                }
                                case 'ENOTFOUND':
                                    // Address cannot be resolved.
                                    return next({ message: "Missing solution " + config.name + " on port " + config.port });
                                default: {
                                    this.logger.error(err);
                                    next(err);
                                }
                            }
                        }
                    });
                    break;
                case "site":
                case "static":
                    this.logger.info(`    Launching solution in 'site' mode`);
                    if (fs_1.default.existsSync(solutionDir + '/client'))
                        handler = express_1.default.static(solutionDir + '/client');
                    else
                        this.logger.error("   " + solutionDir + "/client does not exist");
                    break;
                default: {
                    this.logger.error("   " + solutionDir + " Unknown mode [" + config.mode + "]. Solution will not be loaded.");
                }
                // case "embedded":
                //     handler = express.Router();
                //     break;
            }
            if (handler) {
                router = express_1.default.Router();
                // If there's a doc folder, serve that up.
                if (fs_1.default.existsSync(solutionDir + '/doc')) {
                    router.use(config.path + '/doc', express_1.default.static(solutionDir + '/doc'), (0, serve_index_1.default)(solutionDir + '/doc', { 'icons': true, stylesheet: serveWithStylesheet }));
                }
                // Route the solution's virtual path to the handler.
                if (/static|site/.test(config.mode))
                    router.use(config.path, handler, (0, serve_index_1.default)(solutionDir + '/client', { 'icons': true, stylesheet: serveWithStylesheet }));
                else
                    router.use(config.path, handler);
                // Any paths not found will go to the 404 page...
                router.use((req, res, next) => res.status(404).send((0, errorhandler_1.render404Page)(req, { title: "Failed to route to solution.", message: `Unable to connect to solution at ${req.url}` })));
                router.use(errorhandler_1.ErrorHandler);
                this.logger.info(`    Solution is now exposed at ${config.path}`);
            }
        }
        catch (ex) {
            this.logger.error(`   Failed to initialize ${config.name}! \n${ex.stack || ex.message || JSON.stringify(ex)}`);
        }
        return router;
    }
    discover() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const inventory = {};
            const startTime = new Date().getTime();
            this.logger.info('\n\n\n\n');
            this.logger.info(`Starting solution/extension discovery sweep ${this.blink}...`);
            // Look through each subdir of our parent (i.e. our siblings).
            const folders = (0, utils_1.getDirectories)(this.parentDir);
            for (let i = 0; i < folders.length; i++) {
                const dir = folders[i];
                const solDir = path_1.default.normalize(this.parentDir + dir);
                this.logger.info(`Found '${dir}'...`);
                let pluginJSON = null;
                let configJSON = null;
                let packageJSON = null;
                let isExtension = false;
                let isSolution = false;
                // Ensure that we know what we're looking at, that is it correct, and that 
                // something with this name has not already been discovered during this sweep.
                if (fs_1.default.existsSync(solDir + '/plugin.json') || fs_1.default.existsSync(solDir + '/package.json')) {
                    if (fs_1.default.existsSync(solDir + '/plugin.json'))
                        pluginJSON = JSON.parse(fs_1.default.readFileSync(solDir + '/plugin.json', { encoding: 'utf8' }));
                    else
                        pluginJSON = {};
                    let dssConfig = null;
                    if (fs_1.default.existsSync(solDir + '/dss.config.json'))
                        dssConfig = JSON.parse(fs_1.default.readFileSync(solDir + '/dss.config.json', { encoding: 'utf8' }));
                    if (fs_1.default.existsSync(solDir + '/server/package.json'))
                        packageJSON = JSON.parse(fs_1.default.readFileSync(solDir + '/server/package.json', { encoding: 'utf8' }));
                    else if (fs_1.default.existsSync(solDir + '/package.json'))
                        packageJSON = JSON.parse(fs_1.default.readFileSync(solDir + '/package.json', { encoding: 'utf8' }));
                    else
                        packageJSON = {};
                    if (dssConfig)
                        configJSON = dssConfig;
                    else if (packageJSON["esa-solution-server"])
                        configJSON = JSON.parse(JSON.stringify(packageJSON["esa-solution-server"]));
                    else
                        configJSON = null;
                    if (configJSON) {
                        configJSON.name = configJSON.name || packageJSON.name;
                        configJSON.package = packageJSON;
                        configJSON.version = packageJSON.version;
                        configJSON.id = dir;
                        if (!this.isConfigValid(configJSON, solDir)) {
                            this.logger.error(`   Solution '${configJSON.id}' will not be added`);
                        }
                        else if (inventory.hasOwnProperty(configJSON.id)) {
                            this.logger.error(`   A solution named '${configJSON.id}' was already found in ${inventory[configJSON.id]}`);
                        }
                        else {
                            inventory[configJSON.id] = '/' + dir;
                            isSolution = true;
                        }
                    }
                    else if (inventory.hasOwnProperty(pluginJSON.name)) {
                        this.logger.error(`   An extension named '${pluginJSON.name}' was already found in ${inventory[pluginJSON.name]}`);
                    }
                    else if (pluginJSON.name) {
                        inventory[pluginJSON.name] = '/' + dir;
                        isExtension = true;
                    }
                    else {
                        // NOP
                    }
                }
                else {
                    // NOP
                }
                // Now that we know what we're dealing with, let's act.
                if (isExtension) {
                    const extension = this.findExtensionById(dir);
                    if (!extension) {
                        const extensionConfig = {
                            name: pluginJSON.name,
                            title: pluginJSON.title || ((_a = pluginJSON.configUI) === null || _a === void 0 ? void 0 : _a.displayName) || pluginJSON.name,
                            image: pluginJSON.favicon,
                            id: dir,
                            mode: "extension",
                        };
                        this.extensions.push({
                            name: pluginJSON.name,
                            config: extensionConfig,
                            blink: this.blink, // Revision number.
                        });
                        this.logger.info(`    Extension '${pluginJSON.name}' has been added`);
                    }
                    else if (extension.name !== pluginJSON.name) {
                        this.logger.warn(`    Name has changed from' ${extension.name}'' to '${pluginJSON.name}'`);
                        this.logger.warn(`    This extension will be deleted and reloaded in the next sweep.`);
                    }
                    else {
                        extension.blink = this.blink; // Update blink.
                        this.logger.info(`    Recognized as known extension '${extension.name}'`);
                    }
                }
                else if (isSolution) {
                    // We take note of the name as the AG will see it.
                    configJSON.pluginName = pluginJSON.name;
                    let solution = this.findSolutionById(solDir);
                    // If solution is new, create a traffic router.
                    if (!solution) {
                        const router = this.createRouter(configJSON, solDir);
                        configJSON.status = router ? "ok" : "failedStart";
                        this.solutions.push({
                            name: configJSON.name,
                            router: router,
                            config: configJSON,
                            blink: this.blink,
                        });
                        this.logger.info(`    Solution '${configJSON.name}' has been added`);
                    }
                    else if (this.hasConfigChanged(solution, configJSON)) {
                        const router = this.createRouter(configJSON, solDir);
                        configJSON.status = router ? "ok" : "failedStart";
                        solution.router = router;
                        solution.config = configJSON;
                        solution.blink = this.blink;
                        this.logger.info(`    Solution '${configJSON.name}' has been updated`);
                    }
                    else {
                        this.logger.info(`    Recognized as known solution '${configJSON.name}'`);
                        // Ensure we keep the blink tracked for unchanged plugins.
                        solution.blink = this.blink;
                    }
                }
                else {
                    // It's neither a solution nor extension, so nothing to do.
                }
                // Analyze next folder.
            }
            // If a known solution or extension isn't seen during this blink, remove it.
            this.solutions
                .filter(sol => sol.blink < this.blink) // Blink was not updated...
                .forEach(sol => {
                this.logger.warn(`Folder '${sol.config.id}' or file 'package.json' is no longer present`);
                this.removeSolution(sol);
                this.logger.info(`Solution ${sol.name} has been deleted`);
            });
            this.extensions
                .filter(ext => ext.blink < this.blink) // Blink was not updated...
                .forEach(ext => {
                this.logger.warn(`Folder '${ext.config.id}' or file 'plugin.json' is no longer present`);
                this.removeExtension(ext);
                this.logger.info(`Extension ${ext.name} has been deleted`);
            });
            // Track this sweep as a blink. This is used to discard old solutions 
            // on the next sweep. Prevents breaking things in the middle of a run.
            this.blink++;
            const duration = new Date().getTime() - startTime;
            this.logger.info(`Finished discovery sweep in ${duration}ms.`);
            this.logger.info('\n\n\n\n');
        });
    }
    removeSolution(solution, kill = false) {
        if (!solution)
            return null;
        // Attempt to release all consumed resources.
        // Express 'Router' may require additional review for disposal.
        // The API doesn't clearly define any methods or mechanisms to do so.
        delete solution.router;
        delete solution.config;
        for (var i = 0; i < this.solutions.length; i++)
            if (this.solutions[i] === solution)
                return this.solutions.splice(i, 1)[0];
        return null;
    }
    removeExtension(extension) {
        for (var i = 0; i < this.extensions.length; i++)
            if (this.extensions[i] === extension)
                return this.extensions.splice(i, 1)[0];
        return null;
    }
    findSolutionForPath(path) {
        // Find the solution that handles this path.
        return this.solutions.find(s => s.config.path && path.startsWith(s.config.path + '/'));
    }
    findSolutionById(id) {
        return this.solutions.find(e => e.config.id === id);
    }
    findExtensionById(id) {
        return this.extensions.find(e => e.config.id === id);
    }
    findById(id) {
        return this.extensions.find(e => e.config.id === id) || this.solutions.find(e => e.config.id === id);
    }
    getEcosystemApps(id, reload = false) {
        const solution = this.findById(id);
        // Basic type checking to block all extensions
        if (!solution || !solution.router)
            return null;
        if (!solution.config.ecosystem || reload) {
            const solDir = path_1.default.normalize(__dirname + "/../../" + solution.config.id);
            let ecosystemFile = {};
            if (fs_1.default.existsSync(solDir + '/ecosystem.config.js')) {
                try {
                    ecosystemFile = require('../../' + solution.config.id + '/ecosystem.config.js');
                }
                catch (ex) {
                    // debugger;
                    this.logger.error(ex);
                }
            }
            solution.config.ecosystem = ecosystemFile.apps || ecosystemFile;
        }
        return solution.config.ecosystem;
    }
    getEcosystemFile(name) {
        const solution = this.solutions.find(s => s.config.name === name);
        if (!solution)
            return null;
        return '../../' + solution.config.id + '/ecosystem.config.js';
    }
    requiresRestart(solution) {
        return true; // TODO: Get from ecosystem config function if provided.
    }
    getSolutions() {
        return this.solutions;
    }
    getExtensions() {
        return this.extensions;
    }
    getAll() {
        return [...this.getSolutions(), ...this.getExtensions()];
    }
    start() {
        this.discover();
        this.timer = setInterval(() => this.discover(), this.discoveryInterval);
    }
    stop() {
        clearInterval(this.timer);
    }
}
exports.SolutionFinder = SolutionFinder;
