"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConfigReader = void 0;
const fs_1 = __importDefault(require("fs"));
const js_yaml_1 = __importDefault(require("js-yaml"));
const globals_1 = require("../globals");
const YAML = {
    parse: data => js_yaml_1.default.safeLoad(data),
    stringify: data => js_yaml_1.default.safeDump(data)
};
const INI = {
    parse: data => {
        const result = {};
        // Initally, all key-value pairs belong to the root.
        let section = result;
        data.split(/\r?\n/)
            .map(line => line.trim())
            .filter(line => line !== "" && !/^\#/.test(line))
            .forEach(line => {
            if (line.startsWith('[')) {
                const sectionName = line.substring(1, line.indexOf(']'));
                result[sectionName] = section = {};
                // From now on, key-value pairs go to this section.
            }
            const kvPair = line.split('=');
            const key = kvPair.shift().trim();
            const val = kvPair.join('=');
            section[key] = val;
        });
        return result;
    },
    stringify: data => {
        return "Not yet implemented";
    }
};
const CONF = {
    parse: data => {
        let result = {};
        // Use line trim to clear bad unicode characters.
        // Example Zero Width Space "65279".
        data.split(/[\r\n]/)
            .map(line => line.trim())
            .filter(line => line !== "" && !/^\#/.test(line))
            .forEach(line => {
            const args = line.split(' ');
            const key = args.shift().trim();
            const val = args.join(' ');
            result[key] = val;
        });
        return result;
    },
    stringify: data => {
        return "Not yet implemented";
    }
};
const parsers = {
    json: {
        parse: data => JSON.parse(data.trim()),
        stringify: JSON.stringify
    },
    ini: INI,
    conf: CONF,
    yml: YAML,
    yaml: YAML
};
class ConfigReader {
    constructor(rootDir = "./config/") {
        // Assumption is that we are in '/' by default.
        this.rootDir = "./config/";
        this.watchers = {};
        this.events = {};
        this.rootDir = rootDir;
    }
    emit(eventName, section, data) {
        // Call the subscribed event handlers.
        const event = this.events[eventName] || [];
        event.forEach(handler => handler.call(null, section, data));
    }
    ;
    on(eventName, handler) {
        // Add the subscription (event handler) for this event.
        this.events[eventName] = this.events[eventName] || [];
        this.events[eventName].push(handler);
        return this;
    }
    ;
    parseData(fileName, data, format = undefined) {
        const fileType = format || fileName.split(".").pop();
        if (parsers.hasOwnProperty(fileType))
            return parsers[fileType].parse(data.trim());
        throw new Error("UNKNOWN_FILE_TYPE: " + fileType);
    }
    watchFile(fileName, onChanged) {
        let previousContent = null;
        return fs_1.default.watch(fileName, { encoding: 'utf8' }, event => {
            // File was deleted.
            if (!fs_1.default.existsSync(fileName)) {
                return onChanged(fileName, null);
            }
            // File was updated.
            const currentContent = fs_1.default.readFileSync(fileName, { encoding: 'utf8' });
            if (currentContent !== "" && currentContent !== previousContent) {
                previousContent = currentContent;
                onChanged(fileName, currentContent);
            }
        });
    }
    readFileOrDefault(fileName, fallbackFile = undefined) {
        let fileContent = null;
        if (fs_1.default.existsSync(fileName)) {
            fileContent = fs_1.default.readFileSync(fileName, { encoding: 'utf8' });
        }
        else {
            // Generate '<name>.<type>' from '<name>.default.<type>'.
            const parts = fileName.split(".");
            const ext = parts.pop();
            parts.push("default");
            parts.push(ext);
            // Look for file.default.ext
            // Write file.default.ext to file.ext
            if (fs_1.default.existsSync(parts.join("."))) {
                fileContent = fs_1.default.readFileSync(parts.join("."), { encoding: 'utf8' });
                fs_1.default.writeFileSync(fileName, fileContent);
            }
            else if (fallbackFile) {
                fileContent = fs_1.default.readFileSync(fallbackFile, { encoding: 'utf8' });
                // fs.writeFileSync(fileName, fileContent);
            }
            else {
                // We have nothing to go on...
            }
        }
        return fileContent;
    }
    loadFile(fileName, watchOrFallback = undefined, format = undefined) {
        const log = globals_1.mainLogger;
        try {
            const fallback = typeof watchOrFallback === "string"
                ? watchOrFallback
                : null;
            const watch = typeof watchOrFallback === "boolean"
                ? watchOrFallback
                : true;
            const data = this.readFileOrDefault(fileName, fallback);
            if (!data) {
                log.warn(`File ${fileName} not found.`);
                return null;
            }
            if (watch && !this.watchers[fileName]) {
                this.watchers[fileName] = this.watchFile(fileName, data => {
                    try {
                        const json = this.parseData(fileName, data, format);
                        this.emit('change', fileName, json);
                    }
                    catch (ex) {
                        this.emit('error', fileName + " could not be parsed", ex.message);
                    }
                });
            }
            return this.parseData(fileName, data, format);
        }
        catch (ex) {
            this.emit("error", fileName + " could not be parsed", ex.message);
            return null;
        }
    }
}
exports.ConfigReader = ConfigReader;
;
